from pathlib import Path

import awkward as ak
import numpy as np
import pytest

import pybes3 as p3

data_dir = Path(__file__).parent.parent / "data"

helix_ak = p3.open(data_dir / "test_full_mc_evt_1.dst")[
    "Event/TDstEvent/m_mdcTrackCol"
].array()["m_helix"]

helix_np = ak.flatten(helix_ak, axis=1).to_numpy()


def test_helix():
    # parse_helix
    fields = [
        "x",
        "y",
        "z",
        "r",
        "px",
        "py",
        "pz",
        "pt",
        "p",
        "theta",
        "phi",
        "charge",
        "r_trk",
    ]

    with pytest.warns(DeprecationWarning):
        p_helix_ak1 = p3.parse_helix(helix_ak)
    assert p_helix_ak1.fields == fields

    with pytest.warns(DeprecationWarning):
        p_helix_ak2 = p3.parse_helix(helix_np, library="ak")
    assert p_helix_ak2.fields == fields

    with pytest.warns(DeprecationWarning):
        p_helix_np = p3.parse_helix(helix_np)
    assert list(p_helix_np.keys()) == fields

    # regularize_helix
    with pytest.warns(DeprecationWarning):
        r_helix_ak = p3.regularize_helix(helix_ak)
    assert ak.all(r_helix_ak[..., 0] > 0)
    assert ak.all(r_helix_ak[..., 1] > 0) and ak.all(r_helix_ak[..., 1] < 2 * np.pi)

    with pytest.warns(DeprecationWarning):
        r_helix_np = p3.regularize_helix(helix_np)
    assert np.all(r_helix_np[..., 0] > 0)
    assert np.all(r_helix_np[..., 1] > 0) and np.all(r_helix_np[..., 1] < 2 * np.pi)

    # compute_helix
    with pytest.warns(DeprecationWarning):
        c_helix_ak1 = p3.compute_helix(p_helix_ak1)
    assert ak.all(np.isclose(c_helix_ak1, r_helix_ak))

    with pytest.warns(DeprecationWarning):
        c_helix_ak2 = p3.compute_helix(p_helix_np, library="ak")
    assert ak.all(np.isclose(c_helix_ak2, r_helix_np))

    with pytest.warns(DeprecationWarning):
        c_helix_np = p3.compute_helix(p_helix_np)
    assert np.all(np.isclose(c_helix_np, r_helix_np))


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import awkward as ak
import awkward.contents
import numpy as np


def _extract_index(layout) -> list:
    if isinstance(layout, awkward.contents.ListOffsetArray):
        offsets = layout.offsets.data
        return [offsets[1:] - offsets[:-1]] + _extract_index(layout.content)

    if isinstance(layout, awkward.contents.RegularArray):
        return [layout.size] + _extract_index(layout.content)

    if isinstance(layout, awkward.contents.NumpyArray):
        return []

    if isinstance(layout, awkward.contents.RecordArray):
        return []


def _flat_to_numpy(array) -> np.ndarray:
    """
    Converts a flat awkward array to a numpy array.

    Args:
        array (ak.Array): The input awkward array.

    Returns:
        np.ndarray: The converted numpy array.
    """
    if isinstance(array, (ak.Array, ak.Record)):
        return ak.flatten(array, axis=None).to_numpy()
    else:
        return array


def _recover_shape(array, index) -> ak.Array:
    """
    Recovers the shape of a numpy array based on the provided index.

    Args:
        array (np.ndarray | ak.Array): The input numpy array.
        index (list): The index list that defines the shape.

    Returns:
        ak.Array: The reshaped awkward array.
    """
    res = ak.Array(array)
    for i in index:
        res = ak.unflatten(res, i)
    return res



import awkward as ak
import awkward.contents
import awkward.index
import numpy as np
import uproot
import uproot.behaviors.TBranch
import uproot.extras
import uproot.interpretation
import uproot_custom
from uproot_custom import (
    AsCustom,
    BaseReader,
    BasicTypeReader,
    EmptyReader,
    BaseObjectReader,
    gen_tree_config,
    get_cpp_reader,
    reconstruct_array,
    regularize_object_path,
)

from . import besio_cpp as bcpp


bes3_branch2types = {
    "/Event:TMcEvent/m_mdcMcHitCol": "TMdcMc",
    "/Event:TMcEvent/m_cgemMcHitCol": "TCgemMc",
    "/Event:TMcEvent/m_emcMcHitCol": "TEmcMc",
    "/Event:TMcEvent/m_tofMcHitCol": "TTofMc",
    "/Event:TMcEvent/m_mucMcHitCol": "TMucMc",
    "/Event:TMcEvent/m_mcParticleCol": "TMcParticle",
    "/Event:TDigiEvent/m_mdcDigiCol": "TMdcDigi",
    "/Event:TDigiEvent/m_cgemDigiCol": "TCgemDigi",
    "/Event:TDigiEvent/m_emcDigiCol": "TEmcDigi",
    "/Event:TDigiEvent/m_tofDigiCol": "TTofDigi",
    "/Event:TDigiEvent/m_mucDigiCol": "TMucDigi",
    "/Event:TDigiEvent/m_lumiDigiCol": "TLumiDigi",
    "/Event:TDstEvent/m_mdcTrackCol": "TMdcTrack",
    "/Event:TDstEvent/m_emcTrackCol": "TEmcTrack",
    "/Event:TDstEvent/m_tofTrackCol": "TTofTrack",
    "/Event:TDstEvent/m_mucTrackCol": "TMucTrack",
    "/Event:TDstEvent/m_mdcDedxCol": "TMdcDedx",
    "/Event:TDstEvent/m_extTrackCol": "TExtTrack",
    "/Event:TDstEvent/m_mdcKalTrackCol": "TMdcKalTrack",
    "/Event:TRecEvent/m_recCgemClusterCol": "TRecCgemCluster",
    "/Event:TRecEvent/m_recMdcTrackCol": "TRecMdcTrack",
    "/Event:TRecEvent/m_recMdcHitCol": "TRecMdcHit",
    "/Event:TRecEvent/m_recEmcHitCol": "TRecEmcHit",
    "/Event:TRecEvent/m_recEmcClusterCol": "TRecEmcCluster",
    "/Event:TRecEvent/m_recEmcShowerCol": "TRecEmcShower",
    "/Event:TRecEvent/m_recTofTrackCol": "TRecTofTrack",
    "/Event:TRecEvent/m_recMucTrackCol": "TRecMucTrack",
    "/Event:TRecEvent/m_recMdcDedxCol": "TRecMdcDedx",
    "/Event:TRecEvent/m_recMdcDedxHitCol": "TRecMdcDedxHit",
    "/Event:TRecEvent/m_recExtTrackCol": "TRecExtTrack",
    "/Event:TRecEvent/m_recMdcKalTrackCol": "TRecMdcKalTrack",
    "/Event:TRecEvent/m_recMdcKalHelixSegCol": "TRecMdcKalHelixSeg",
    "/Event:TRecEvent/m_recEvTimeCol": "TRecEvTime",
    "/Event:TRecEvent/m_recZddChannelCol": "TRecZddChannel",
    "/Event:TEvtRecObject/m_evtRecTrackCol": "TEvtRecTrack",
    "/Event:TEvtRecObject/m_evtRecVeeVertexCol": "TEvtRecVeeVertex",
    "/Event:TEvtRecObject/m_evtRecPi0Col": "TEvtRecPi0",
    "/Event:TEvtRecObject/m_evtRecEtaToGGCol": "TEvtRecEtaToGG",
    "/Event:TEvtRecObject/m_evtRecDTagCol": "TEvtRecDTag",
    "/Event:THltEvent/m_hltRawCol": "THltRaw",
    "/Event:EventNavigator/m_mcMdcMcHits": "map<int,int>",
    "/Event:EventNavigator/m_mcMdcTracks": "map<int,int>",
    "/Event:EventNavigator/m_mcEmcMcHits": "map<int,int>",
    "/Event:EventNavigator/m_mcEmcRecShowers": "map<int,int>",
}


class Bes3TObjArrayReader(BaseReader):
    @classmethod
    def priority(cls):
        return 50

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name: str,
        cls_streamer_info: dict,
        all_streamer_info: dict,
        item_path: str,
        called_from_top: bool,
    ):
        if top_type_name != "TObjArray":
            return None

        item_path = item_path.replace(".TObjArray*", "")
        obj_typename = bes3_branch2types.get(item_path)
        if obj_typename is None:
            return None

        if obj_typename not in all_streamer_info:
            return {
                "reader": cls,
                "name": cls_streamer_info["fName"],
                "element_reader": {
                    "reader": EmptyReader,
                    "name": obj_typename,
                },
            }

        sub_reader_config = []
        for s in all_streamer_info[obj_typename]:
            sub_reader_config.append(
                gen_tree_config(
                    cls_streamer_info=s,
                    all_streamer_info=all_streamer_info,
                    item_path=f"{item_path}.{obj_typename}",
                )
            )

        return {
            "reader": cls,
            "name": cls_streamer_info["fName"],
            "element_reader": {
                "reader": BaseObjectReader,
                "name": obj_typename,
                "sub_readers": sub_reader_config,
            },
        }

    @classmethod
    def get_cpp_reader(cls, reader_config: dict):
        if reader_config["reader"] != cls:
            return None

        element_reader_config = reader_config["element_reader"]
        element_reader = get_cpp_reader(element_reader_config)

        return bcpp.Bes3TObjArrayReader(reader_config["name"], element_reader)

    @classmethod
    def reconstruct_array(cls, raw_data, reader_config: dict):
        if reader_config["reader"] != cls:
            return None

        offsets, element_raw_data = raw_data
        element_reader_config = reader_config["element_reader"]
        element_data = reconstruct_array(
            element_raw_data,
            element_reader_config,
        )

        return awkward.contents.ListOffsetArray(
            awkward.index.Index64(offsets),
            element_data,
        )


class Bes3CgemClusterColReader(BaseReader):
    @classmethod
    def priority(cls):
        return 55

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
        called_from_top: bool,
    ):
        item_path = item_path.replace(".TObjArray*", "")
        if item_path != "/Event:TRecEvent/m_recCgemClusterCol":
            return None

        if all_streamer_info.get("TCgemCluster") is not None:
            return None  # Let Bes3TObjArrayReader handle it

        return {
            "reader": cls,
            "name": cls_streamer_info["fName"],
        }

    @classmethod
    def get_cpp_reader(cls, tree_config):
        if tree_config["reader"] != cls:
            return None

        return bcpp.Bes3CgemClusterColReader(tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != cls:
            return None

        offsets = raw_data.pop("offsets")

        record_contents = []
        for k, v in raw_data.items():
            tmp_content = awkward.contents.NumpyArray(v)

            if k == "m_clusterFlag":
                tmp_content = awkward.contents.RegularArray(tmp_content, 2)

            if k == "m_stripID":
                tmp_content = awkward.contents.RegularArray(tmp_content, 2)
                tmp_content = awkward.contents.RegularArray(tmp_content, 2)

            record_contents.append(tmp_content)

        return awkward.contents.ListOffsetArray(
            awkward.index.Index64(offsets),
            awkward.contents.RecordArray(record_contents, list(raw_data.keys())),
        )


class Bes3SymMatrixArrayReader(BaseReader):
    target_items = {
        "/Event:TDstEvent/m_mdcTrackCol.TMdcTrack.m_err",
        "/Event:TDstEvent/m_emcTrackCol.TEmcTrack.m_err",
        "/Event:TDstEvent/m_extTrackCol.TExtTrack.myTof1ErrorMatrix",
        "/Event:TDstEvent/m_extTrackCol.TExtTrack.myTof2ErrorMatrix",
        "/Event:TDstEvent/m_extTrackCol.TExtTrack.myEmcErrorMatrix",
        "/Event:TDstEvent/m_extTrackCol.TExtTrack.myMucErrorMatrix",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_zerror",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_zerror_e",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_zerror_mu",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_zerror_k",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_zerror_p",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_ferror",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_ferror_e",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_ferror_mu",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_ferror_k",
        "/Event:TDstEvent/m_mdcKalTrackCol.TMdcKalTrack.m_ferror_p",
        "/Event:TEvtRecObject/m_evtRecVeeVertexCol.TEvtRecVeeVertex.m_Ew",
        "/Event:TEvtRecObject/m_evtRecPrimaryVertex.m_Evtx",  # TODO: use BES3 interpretation
        "/Event:TRecEvent/m_recMdcTrackCol.TRecMdcTrack.m_err",
        "/Event:TRecEvent/m_recEmcShowerCol.TRecEmcShower.m_err",
        "/Event:TRecEvent/m_recMdcKalTrackCol.TRecMdcKalTrack.m_terror",
    }

    @classmethod
    def priority(cls):
        return 40

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
        called_from_top: bool,
    ):
        if item_path not in Bes3SymMatrixArrayReader.target_items:
            return None

        fArrayDim = cls_streamer_info["fArrayDim"]
        fMaxIndex = cls_streamer_info["fMaxIndex"]
        ctype = BasicTypeReader.typenames[top_type_name]

        flat_size = np.prod(fMaxIndex[:fArrayDim])
        assert flat_size > 0, f"flatten_size should be greater than 0, but got {flat_size}"

        full_dim = int((np.sqrt(1 + 8 * flat_size) - 1) / 2)
        return {
            "reader": cls,
            "name": cls_streamer_info["fName"],
            "ctype": ctype,
            "flat_size": flat_size,
            "full_dim": full_dim,
        }

    @classmethod
    def get_cpp_reader(cls, tree_config):
        if tree_config["reader"] != cls:
            return None

        ctype = tree_config["ctype"]
        assert ctype == "d", "Only double precision symmetric matrix is supported."

        return bcpp.Bes3SymMatrixArrayReader(
            tree_config["name"],
            tree_config["flat_size"],
            tree_config["full_dim"],
        )

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != cls:
            return None

        full_dim = tree_config["full_dim"]
        return awkward.contents.NumpyArray(raw_data.reshape(-1, full_dim, full_dim))


uproot_custom.registered_readers |= {
    Bes3TObjArrayReader,
    Bes3SymMatrixArrayReader,
    Bes3CgemClusterColReader,
}


##########################################################################################
#                                     Array Preprocess
##########################################################################################


#############################################
# TDigiEvent
#############################################
def process_digi_subbranch(org_arr: ak.Array) -> ak.Array:
    """
    Processes the `TRawData` subbranch of the input awkward array and returns a new array with the subbranch fields
    merged into the top level.

    Parameters:
        org_arr (ak.Array): The input awkward array containing the `TRawData` subbranch.

    Returns:
        A new awkward array with the fields of `TRawData` merged into the top level.

    Raises:
        AssertionError: If `TRawData` is not found in the input array fields.
    """
    if not org_arr.fields:
        assert ak.count(org_arr) == 0, "Input array is empty but has no fields"
        return org_arr

    assert "TRawData" in org_arr.fields, "TRawData not found in the input array"

    fields = {}
    for field_name in org_arr.fields:
        if field_name == "TRawData":
            for raw_field_name in org_arr[field_name].fields:
                fields[raw_field_name] = org_arr[field_name][raw_field_name]
        else:
            fields[field_name] = org_arr[field_name]

    return ak.zip(fields)


#############################################
# Main function
#############################################
def preprocess_subbranch(full_branch_path: str, org_arr: ak.Array) -> ak.Array:
    full_branch_path = full_branch_path.replace("/Event:", "")
    evt_name, subbranch_name = full_branch_path.split("/")

    if evt_name == "TDigiEvent" and subbranch_name != "m_fromMc":
        return process_digi_subbranch(org_arr)

    # Default return
    return org_arr


class Bes3Interpretation(AsCustom):
    """
    Custom interpretation for Bes3 data.
    """

    target_branches: set[str] = set(bes3_branch2types.keys())

    def __init__(self, branch, context, simplify):
        super().__init__(branch, context, simplify)
        self._typename = bes3_branch2types[regularize_object_path(branch.object_path)]

    def basket_array(
        self,
        data,
        byte_offsets,
        basket,
        branch,
        context,
        cursor_offset,
        library,
        interp_options,
    ):
        raw_ak_layout = super().basket_array(
            data,
            byte_offsets,
            basket,
            branch,
            context,
            cursor_offset,
            library,
            interp_options,
        )
        raw_ak_arr = ak.Array(raw_ak_layout)

        # preprocess awkward array and return
        full_branch_path = regularize_object_path(branch.object_path)
        return preprocess_subbranch(full_branch_path, raw_ak_arr)

    @property
    def typename(self) -> str:
        """
        The typename of the interpretation.
        """
        return self._typename

    def __repr__(self) -> str:
        """
        The string representation of the interpretation.
        """
        return f"AsBes3(TObjArray[{self.typename}])"


##########################################################################################
#                                       Wrappers
##########################################################################################
_is_uproot_interpretation_of_wrapped = False

_uproot_interpretation_of = uproot.interpretation.identify.interpretation_of


def bes_interpretation_of(
    branch: uproot.behaviors.TBranch.TBranch, context: dict, simplify: bool = True
) -> uproot.interpretation.Interpretation:
    if not hasattr(branch, "parent"):
        return _uproot_interpretation_of(branch, context, simplify)

    if Bes3Interpretation.match_branch(branch, context, simplify):
        return Bes3Interpretation(branch, context, simplify)

    return _uproot_interpretation_of(branch, context, simplify)


def wrap_uproot_interpretation():
    global _is_uproot_interpretation_of_wrapped
    if not _is_uproot_interpretation_of_wrapped:
        _is_uproot_interpretation_of_wrapped = True
        uproot.interpretation.identify.interpretation_of = bes_interpretation_of


def wrap_uproot():
    """
    Wraps the uproot functions to use the BES interpretation.
    """
    wrap_uproot_interpretation()



from pathlib import Path

import awkward as ak
import numpy as np
import pytest
import vector

import pybes3 as p3

data_dir = Path(__file__).parent.parent / "data"
mdc_trk = p3.open(data_dir / "test_full_mc_evt_1.dst")["Event/TDstEvent/m_mdcTrackCol"].array()

raw_helix_arr = mdc_trk["m_helix"]
raw_helix_err_arr = mdc_trk["m_err"]

flat_helix_arr = ak.flatten(raw_helix_arr)
flat_helix_err_arr = ak.flatten(raw_helix_err_arr)


def _make_arr_pivot(base_arr, x, y, z):
    x = ak.flatten(ak.ones_like(base_arr) * x, axis=None)
    y = ak.flatten(ak.ones_like(base_arr) * y, axis=None)
    z = ak.flatten(ak.ones_like(base_arr) * z, axis=None)


@pytest.mark.parametrize("error", [None, flat_helix_err_arr])
def test_helix_obj_1(error):
    """Test initialization, momentum, position, charge, and pivot of helix_obj."""
    raw_pivot = (0, 0, 0)

    for i in range(len(flat_helix_arr)):
        dr, phi0, kappa, dz, tanl = flat_helix_arr[i]
        np_error = error[i].to_numpy() if error is not None else None

        # case 1
        h1 = p3.helix_obj(
            dr,
            phi0,
            kappa,
            dz,
            tanl,
            error=np_error,
            pivot=raw_pivot,
        )

        # case 2
        h2 = p3.helix_obj(
            params=(dr, phi0, kappa, dz, tanl),
            error=np_error,
            pivot=raw_pivot,
        )

        momentum = h1.momentum
        position = h1.position

        # case 3-1: momentum and position are both vector
        h3_1 = p3.helix_obj(
            momentum=momentum,
            position=position,
            charge=h1.charge,
            error=np_error,
            pivot=raw_pivot,
        )

        # case 3-2: momentum as tuple, position as vector
        h3_2 = p3.helix_obj(
            momentum=(momentum.px, momentum.py, momentum.pz),
            position=position,
            charge=h1.charge,
            error=np_error,
            pivot=raw_pivot,
        )

        # case 3-3: momentum as vector, position as tuple
        h3_3 = p3.helix_obj(
            momentum=momentum,
            position=(position.x, position.y, position.z),
            charge=h1.charge,
            error=np_error,
            pivot=raw_pivot,
        )

        # case 3-4: momentum and position as tuple
        h3_4 = p3.helix_obj(
            momentum=(momentum.px, momentum.py, momentum.pz),
            position=(position.x, position.y, position.z),
            charge=h1.charge,
            error=np_error,
            pivot=raw_pivot,
        )

        for h in [h1, h2, h3_1, h3_2, h3_3, h3_4]:
            assert h.dr == pytest.approx(dr, rel=1e-6)
            assert h.phi0 == pytest.approx(phi0, rel=1e-6)
            assert h.kappa == pytest.approx(kappa, rel=1e-6)
            assert h.dz == pytest.approx(dz, rel=1e-6)
            assert h.tanl == pytest.approx(tanl, rel=1e-6)
            assert h.pivot.x == raw_pivot[0]
            assert h.pivot.y == raw_pivot[1]
            assert h.pivot.z == raw_pivot[2]

            if error is not None:
                assert h.error is not None
                assert np.all(h.error == np_error)
            else:
                assert h.error is None


def test_helix_obj_2():
    """Test change_pivot and isclose methods of helix_obj."""
    new_pivots = [
        (10, 10, 10),
        (10, -10, 10),
        (-10, -10, 10),
        (-10, 10, 10),
        (10, 10, 10),  # --- z > 0 / z < 0 ---
        (10, 10, -10),
        (10, -10, -10),
        (-10, -10, -10),
        (-10, 10, -10),
    ]

    for i in range(len(flat_helix_arr)):
        dr, phi0, kappa, dz, tanl = flat_helix_arr[i]
        helix_err = flat_helix_err_arr[i].to_numpy()

        h1_raw = p3.helix_obj(params=(dr, phi0, kappa, dz, tanl), error=helix_err)
        h2_raw = p3.helix_obj(params=(dr, phi0, kappa, dz, tanl), error=helix_err)
        h3_raw = p3.helix_obj(params=(dr, phi0, kappa, dz, tanl))
        h4_raw = p3.helix_obj(params=(dr, phi0, kappa, dz, tanl))

        h1 = h1_raw
        h2 = h2_raw
        h3 = h3_raw
        h4 = h4_raw

        for new_pivot in new_pivots:
            h1 = h1.change_pivot(*new_pivot)
            h2 = h2.change_pivot(*new_pivot)

            vec_new_pivot = vector.obj(x=new_pivot[0], y=new_pivot[1], z=new_pivot[2])
            h3 = h3.change_pivot(vec_new_pivot)
            h4 = h4.change_pivot(vec_new_pivot)

        # changed pivot, should be still close because of the automatic transformation
        assert h1.isclose(h1_raw)
        assert h2.isclose(h2_raw)
        assert h3.isclose(h3_raw)
        assert h4.isclose(h4_raw)

        # check strict criteria
        assert not h1.isclose(h1_raw, rtol=0, atol=0)

    # test with-error and no-error helix comparison
    with pytest.warns(UserWarning, match="Ignoring error matrix for isclose check."):
        h1_raw.isclose(h3_raw)


@pytest.mark.parametrize(
    "init_pivot",
    [
        (0, 0, 0),
        vector.obj(x=0, y=0, z=0),
        vector.obj(rho=0, phi=0, z=0),
    ],
)
@pytest.mark.parametrize(
    "raw_pivot",
    [
        (0, 0, 0),
        (vector.obj(x=0, y=0, z=0),),
    ],
)
@pytest.mark.parametrize(
    "new_pivot",
    [
        (10, 10, 10),
        (vector.obj(x=10, y=10, z=10),),
    ],
)
def test_helix_obj_3(init_pivot, raw_pivot, new_pivot):
    """Test different pivot arguments."""
    for i in range(len(flat_helix_arr)):
        h1 = p3.helix_obj(params=flat_helix_arr[i].tolist(), pivot=init_pivot)
        assert h1.change_pivot(*new_pivot).change_pivot(*raw_pivot).isclose(h1)


def test_helix_awk_1():
    """Test helix_awk, change_pivot, and isclose methods."""
    raw_pivot = ak.zip(
        {
            "x": ak.zeros_like(raw_helix_arr[..., 0]),
            "y": ak.zeros_like(raw_helix_arr[..., 0]),
            "z": ak.zeros_like(raw_helix_arr[..., 0]),
        },
        with_name="Vector3D",
    )

    new_pivot = ak.zip(
        {
            "x": ak.ones_like(raw_helix_arr[..., 0]) * 10,
            "y": ak.ones_like(raw_helix_arr[..., 0]) * 10,
            "z": ak.ones_like(raw_helix_arr[..., 0]) * 10,
        },
        with_name="Vector3D",
    )

    # directly constructing the helix array
    h0 = ak.Array(
        {
            "dr": raw_helix_arr[..., 0],
            "phi0": raw_helix_arr[..., 1],
            "kappa": raw_helix_arr[..., 2],
            "dz": raw_helix_arr[..., 3],
            "tanl": raw_helix_arr[..., 4],
            "pivot": raw_pivot,
            "error": raw_helix_err_arr,
        },
        with_name="Bes3Helix",
    )

    assert ak.all(h0.change_pivot(new_pivot).change_pivot(raw_pivot).isclose(h0))

    # test helix_awk and change_pivot

    # case 1-1: positional helix and error arguments
    h1_1 = p3.helix_awk(raw_helix_arr, raw_helix_err_arr)
    assert ak.all(h0.isclose(h1_1))
    assert ak.all(h1_1.change_pivot(new_pivot).change_pivot(raw_pivot).isclose(h1_1))

    # case 1-2: keyword arguments helix, error
    h1_2 = p3.helix_awk(helix=raw_helix_arr, error=raw_helix_err_arr)
    assert ak.all(h0.isclose(h1_2))

    # case 2: keyword arguments dr, phi0, kappa, dz, tanl
    h2 = p3.helix_awk(
        dr=raw_helix_arr[..., 0],
        phi0=raw_helix_arr[..., 1],
        kappa=raw_helix_arr[..., 2],
        dz=raw_helix_arr[..., 3],
        tanl=raw_helix_arr[..., 4],
        error=raw_helix_err_arr,
    )
    assert ak.all(h0.isclose(h2))

    # case 3: keyword arguments momentum, position, charge
    momentum = h0.momentum
    position = h0.position
    charge = h0.charge
    h3 = p3.helix_awk(
        momentum=momentum,
        position=position,
        charge=charge,
        error=raw_helix_err_arr,
    )
    assert ak.all(h0.isclose(h3))

    # test isclose method, no-error and with-error helix comparison
    h4 = p3.helix_awk(raw_helix_arr)
    with pytest.warns(UserWarning, match="Ignoring error matrix for isclose check."):
        assert ak.all(h1_1.isclose(h4))

    # check strict criteria
    assert not ak.all(
        h0.change_pivot(new_pivot).change_pivot(raw_pivot).isclose(h0, atol=0, rtol=0)
    )

    # changed pivot
    h4 = p3.helix_awk(helix=raw_helix_arr, error=raw_helix_err_arr).change_pivot(new_pivot)
    assert ak.all(h4.isclose(h0))

    # different pivot arguments
    assert ak.all(h0.change_pivot(new_pivot).change_pivot(0, 0, 0).isclose(h0))


@pytest.mark.parametrize(
    "init_pivot",
    [
        ak.zip(
            {
                "x": ak.zeros_like(raw_helix_arr[..., 0]),
                "y": ak.zeros_like(raw_helix_arr[..., 0]),
                "z": ak.zeros_like(raw_helix_arr[..., 0]),
            },
            with_name="Vector3D",
        ),
        (0, 0, 0),
    ],
)
@pytest.mark.parametrize(
    "raw_pivot",
    [
        (
            ak.zip(
                {
                    "x": ak.zeros_like(raw_helix_arr[..., 0]),
                    "y": ak.zeros_like(raw_helix_arr[..., 0]),
                    "z": ak.zeros_like(raw_helix_arr[..., 0]),
                },
                with_name="Vector3D",
            ),
        ),
        (0, 0, 0),
    ],
)
@pytest.mark.parametrize(
    "new_pivot",
    [
        (
            ak.zip(
                {
                    "x": ak.ones_like(raw_helix_arr[..., 0]) * 10,
                    "y": ak.ones_like(raw_helix_arr[..., 0]) * 10,
                    "z": ak.ones_like(raw_helix_arr[..., 0]) * 10,
                },
                with_name="Vector3D",
            ),
        ),
        (10, 10, 10),
    ],
)
def test_helix_awk_2(init_pivot, raw_pivot, new_pivot):
    """Test helix_awk with different pivot arguments."""
    h = p3.helix_awk(
        helix=raw_helix_arr,
        error=raw_helix_err_arr,
        pivot=init_pivot,
    )
    assert ak.all(h.change_pivot(*new_pivot).change_pivot(*raw_pivot).isclose(h))


@pytest.mark.parametrize(
    "new_pivot",
    [
        (10, 10, 10),
        (vector.obj(x=10, y=10, z=10),),
    ],
)
def test_HelixAwkwardRecord_1(new_pivot):
    """Test HelixAwkwardRecord class with 1 track."""
    helix_arr = p3.helix_awk(helix=raw_helix_arr, error=raw_helix_err_arr)
    helix_rec = helix_arr[0, 0]
    helix_obj = p3.helix_obj(
        dr=helix_rec.dr,
        phi0=helix_rec.phi0,
        kappa=helix_rec.kappa,
        dz=helix_rec.dz,
        tanl=helix_rec.tanl,
        pivot=(helix_rec.pivot.x, helix_rec.pivot.y, helix_rec.pivot.z),
        error=helix_rec.error,
    )

    # test isclose
    assert helix_rec.change_pivot(*new_pivot).change_pivot(0, 0, 0).isclose(helix_rec)
    assert helix_rec.change_pivot(*new_pivot).isclose(helix_rec)

    # test attributes
    assert isinstance(helix_rec.position, ak.Record)
    assert isinstance(helix_rec.momentum, ak.Record)
    assert helix_rec.position.isclose(helix_obj.position)
    assert helix_rec.momentum.isclose(helix_obj.momentum)
    assert helix_rec.charge == helix_obj.charge
    assert helix_rec.radius == pytest.approx(helix_obj.radius, rel=1e-6)


@pytest.mark.parametrize(
    "new_pivot",
    [
        (10, 10, 10),
        (vector.obj(x=10, y=10, z=10),),
    ],
)
def test_helix_awk_3(new_pivot):
    """Test HelixAwkwardArray slices."""

    helix_arr = p3.helix_awk(helix=raw_helix_arr, error=raw_helix_err_arr)
    helix_rec = helix_arr[0]

    # test isclose
    assert ak.all(helix_rec.change_pivot(*new_pivot).change_pivot(0, 0, 0).isclose(helix_rec))
    assert ak.all(helix_rec.change_pivot(*new_pivot).isclose(helix_rec))

    # test attributes
    assert isinstance(helix_rec.position, ak.Array)
    assert isinstance(helix_rec.momentum, ak.Array)
    assert ak.all(helix_rec.position.isclose(helix_arr.position[0]))
    assert ak.all(helix_rec.momentum.isclose(helix_arr.momentum[0]))
    assert ak.all(helix_rec.charge == helix_arr.charge[0])
    assert ak.all(np.isclose(helix_rec.radius, helix_arr.radius[0]))


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



from typing import Optional

import numba as nb
import numpy as np

from ..typing import BoolLike, IntLike

DIGI_MDC_FLAG = np.uint32(0x10)
DIGI_TOF_FLAG = np.uint32(0x20)
DIGI_EMC_FLAG = np.uint32(0x30)
DIGI_MUC_FLAG = np.uint32(0x40)
DIGI_HLT_FLAG = np.uint32(0x50)
DIGI_CGEM_FLAG = np.uint32(0x60)
DIGI_MRPC_FLAG = np.uint32(0x70)
DIGI_FLAG_OFFSET = np.uint32(24)
DIGI_FLAG_MASK = np.uint32(0xFF000000)

# MDC
DIGI_MDC_WIRETYPE_OFFSET = np.uint32(15)
DIGI_MDC_WIRETYPE_MASK = np.uint32(0x00008000)
DIGI_MDC_LAYER_OFFSET = np.uint32(9)
DIGI_MDC_LAYER_MASK = np.uint32(0x00007E00)
DIGI_MDC_WIRE_OFFSET = np.uint32(0)
DIGI_MDC_WIRE_MASK = np.uint32(0x000001FF)
DIGI_MDC_STEREO_WIRE = np.uint32(1)

# TOF
DIGI_TOF_PART_OFFSET = np.uint32(14)
DIGI_TOF_PART_MASK = np.uint32(0x0000C000)
DIGI_TOF_END_OFFSET = np.uint32(0)
DIGI_TOF_END_MASK = np.uint32(0x00000001)

DIGI_TOF_SCINT_LAYER_OFFSET = np.uint32(8)
DIGI_TOF_SCINT_LAYER_MASK = np.uint32(0x00000100)
DIGI_TOF_SCINT_PHI_OFFSET = np.uint32(1)
DIGI_TOF_SCINT_PHI_MASK = np.uint32(0x000000FE)

DIGI_TOF_MRPC_ENDCAP_OFFSET = np.uint32(11)
DIGI_TOF_MRPC_ENDCAP_MASK = np.uint32(0x00000800)
DIGI_TOF_MRPC_MODULE_OFFSET = np.uint32(5)
DIGI_TOF_MRPC_MODULE_MASK = np.uint32(0x000007E0)
DIGI_TOF_MRPC_STRIP_OFFSET = np.uint32(1)
DIGI_TOF_MRPC_STRIP_MASK = np.uint32(0x0000001E)

# EMC
DIGI_EMC_MODULE_OFFSET = np.uint32(16)
DIGI_EMC_MODULE_MASK = np.uint32(0x000F0000)
DIGI_EMC_THETA_OFFSET = np.uint32(8)
DIGI_EMC_THETA_MASK = np.uint32(0x00003F00)
DIGI_EMC_PHI_OFFSET = np.uint32(0)
DIGI_EMC_PHI_MASK = np.uint32(0x000000FF)

# MUC
DIGI_MUC_PART_OFFSET = np.uint32(16)
DIGI_MUC_PART_MASK = np.uint32(0x000F0000)
DIGI_MUC_SEGMENT_OFFSET = np.uint32(12)
DIGI_MUC_SEGMENT_MASK = np.uint32(0x0000F000)
DIGI_MUC_LAYER_OFFSET = np.uint32(8)
DIGI_MUC_LAYER_MASK = np.uint32(0x00000F00)
DIGI_MUC_CHANNEL_OFFSET = np.uint32(0)
DIGI_MUC_CHANNEL_MASK = np.uint32(0x000000FF)

# CGEM
DIGI_CGEM_STRIP_OFFSET = np.uint32(7)
DIGI_CGEM_STRIP_MASK = np.uint32(0x0007FF80)
DIGI_CGEM_STRIPTYPE_OFFSET = np.uint32(6)
DIGI_CGEM_STRIPTYPE_MASK = np.uint32(0x00000040)
DIGI_CGEM_SHEET_OFFSET = np.uint32(3)
DIGI_CGEM_SHEET_MASK = np.uint32(0x00000038)
DIGI_CGEM_LAYER_OFFSET = np.uint32(0)
DIGI_CGEM_LAYER_MASK = np.uint32(0x00000007)
DIGI_CGEM_XSTRIP = np.uint32(0)


###############################################################################
#                                     MDC                                     #
###############################################################################
@nb.vectorize(cache=True)
def check_mdc_id(mdc_digi_id: IntLike) -> BoolLike:
    """
    Check if the MDC digi ID is valid.

    Parameters:
        mdc_digi_id: The MDC digi ID array or value.

    Returns:
        Whether the digi ID is valid.
    """
    return (mdc_digi_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_MDC_FLAG


@nb.vectorize(cache=True)
def mdc_id_to_wire(mdc_digi_id: IntLike) -> IntLike:
    """
    Convert MDC digi ID to wire number.

    Parameters:
        mdc_digi_id: MDC digi ID array or value.

    Returns:
        The wire number.
    """
    return np.uint16((mdc_digi_id & DIGI_MDC_WIRE_MASK) >> DIGI_MDC_WIRE_OFFSET)


@nb.vectorize(cache=True)
def mdc_id_to_layer(mdc_digi_id: IntLike) -> IntLike:
    """
    Convert the MDC digi ID to the layer number.

    Parameters:
        mdc_digi_id: The MDC digi ID array or value.

    Returns:
        The layer number.
    """
    return np.uint8((mdc_digi_id & DIGI_MDC_LAYER_MASK) >> DIGI_MDC_LAYER_OFFSET)


@nb.vectorize(cache=True)
def mdc_id_to_is_stereo(mdc_digi_id: IntLike) -> BoolLike:
    """
    Convert the MDC digi ID to whether it is a stereo wire.

    Parameters:
        mdc_digi_id: The MDC digi ID array or value.

    Returns:
        Whether the wire is a stereo wire.
    """
    return (
        mdc_digi_id & DIGI_MDC_WIRETYPE_MASK
    ) >> DIGI_MDC_WIRETYPE_OFFSET == DIGI_MDC_STEREO_WIRE


@nb.vectorize(cache=True)
def get_mdc_digi_id(
    wire: IntLike,
    layer: IntLike,
    wire_type: IntLike,
) -> IntLike:
    """
    Generate MDC digi ID based on the wire number, layer number, and wire type.

    Parameters:
        wire: The wire number.
        layer: The layer number.
        wire_type: The wire type.

    Returns:
        The MDC digi ID.
    """
    return np.uint32(
        ((wire << DIGI_MDC_WIRE_OFFSET) & DIGI_MDC_WIRE_MASK)
        | ((layer << DIGI_MDC_LAYER_OFFSET) & DIGI_MDC_LAYER_MASK)
        | ((wire_type << DIGI_MDC_WIRETYPE_OFFSET) & DIGI_MDC_WIRETYPE_MASK)
        | (DIGI_MDC_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                     TOF                                     #
###############################################################################
@nb.vectorize(cache=True)
def check_tof_id(tof_digi_id: IntLike) -> BoolLike:
    """
    Check if the TOF digi ID is valid.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.

    Returns:
        Whether the digi ID is valid.
    """
    return (tof_digi_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_TOF_FLAG


@nb.vectorize(cache=True)
def tof_id_to_part(tof_digi_id: IntLike) -> IntLike:
    """
    Convert TOF digi ID to part number. 0, 1, 2 for scintillator endcap0/barrel/endcap1,
    3, 4 for MRPC endcap0/endcap1.

    Parameters:
        tof_digi_id: TOF digi ID array or value.

    Returns:
        The part number.
    """
    part = (tof_digi_id & DIGI_TOF_PART_MASK) >> DIGI_TOF_PART_OFFSET
    if part == 3:  # += MRPC endcap number
        part += (tof_digi_id & DIGI_TOF_MRPC_ENDCAP_MASK) >> DIGI_TOF_MRPC_ENDCAP_OFFSET
    return np.uint8(part)


@nb.vectorize(cache=True)
def tof_id_to_end(tof_digi_id: IntLike) -> IntLike:
    """
    Convert the TOF digi ID to the readout end number.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.

    Returns:
        The readout end number.
    """
    return np.uint8(tof_digi_id % 2)


@nb.vectorize(cache=True)
def _tof_id_to_layer_or_module_1(tof_digi_id: IntLike) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator layer or MRPC module number.
    No part number is provided, so it will be calculated based on the TOF digi ID.

    This function is used by `tof_id_to_layerOrModule` when part number is not provided.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.

    Returns:
        The scintillator layer or MRPC module number.
    """
    part = tof_id_to_part(tof_digi_id)
    if part < 3:
        res = (tof_digi_id & DIGI_TOF_SCINT_LAYER_MASK) >> DIGI_TOF_SCINT_LAYER_OFFSET
    else:
        res = (tof_digi_id & DIGI_TOF_MRPC_MODULE_MASK) >> DIGI_TOF_MRPC_MODULE_OFFSET
    return np.uint8(res)


@nb.vectorize(cache=True)
def _tof_id_to_layer_or_module_2(tof_digi_id: IntLike, part: IntLike) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator layer or MRPC module number.

    This function is used by `tof_id_to_layerOrModule` when part number is provided.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.
        part: The part number.

    Returns:
        The scintillator layer or MRPC module number based on the part number.
    """
    if part < 3:
        res = (tof_digi_id & DIGI_TOF_SCINT_LAYER_MASK) >> DIGI_TOF_SCINT_LAYER_OFFSET
    else:
        res = (tof_digi_id & DIGI_TOF_MRPC_MODULE_MASK) >> DIGI_TOF_MRPC_MODULE_OFFSET
    return np.uint8(res)


def tof_id_to_layer_or_module(
    tof_digi_id: IntLike,
    part: Optional[IntLike] = None,
) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator layer or MRPC module number.
    If `part < 3`, it is scintillator and the return value is layer number. Otherwise, it is
    MRPC and the return value is module number.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.
        part: The part number. If not provided, it will be calculated based on the TOF digi ID.

    Returns:
        The scintillator layer or MRPC module number.
    """
    if part is None:
        return _tof_id_to_layer_or_module_1(tof_digi_id)
    else:
        return _tof_id_to_layer_or_module_2(tof_digi_id, part)


@nb.vectorize(cache=True)
def _tof_id_to_phi_or_strip_1(tof_digi_id: IntLike) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator phi or MRPC strip number.
    No part number is provided, so it will be calculated based on the TOF digi ID.

    This function is used by `tof_id_to_phiOrStrip` when part number is not provided.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.

    Returns:
        The scintillator phi or MRPC strip number.
    """
    part = tof_id_to_part(tof_digi_id)
    if part < 3:
        res = (tof_digi_id & DIGI_TOF_SCINT_PHI_MASK) >> DIGI_TOF_SCINT_PHI_OFFSET
    else:
        res = (tof_digi_id & DIGI_TOF_MRPC_STRIP_MASK) >> DIGI_TOF_MRPC_STRIP_OFFSET
    return np.uint8(res)


@nb.vectorize(cache=True)
def _tof_id_to_phi_or_strip_2(tof_digi_id: IntLike, part: IntLike) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator phi or MRPC strip number.

    This function is used by `tof_id_to_phiOrStrip` when part number is provided.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.
        part: The part number.

    Returns:
        The scintillator phi or MRPC strip number based on the part number.
    """
    if part < 3:
        res = (tof_digi_id & DIGI_TOF_SCINT_PHI_MASK) >> DIGI_TOF_SCINT_PHI_OFFSET
    else:
        res = (tof_digi_id & DIGI_TOF_MRPC_STRIP_MASK) >> DIGI_TOF_MRPC_STRIP_OFFSET
    return np.uint8(res)


def tof_id_to_phi_or_strip(
    tof_digi_id: IntLike,
    part: Optional[IntLike] = None,
) -> IntLike:
    """
    Convert the TOF digi ID to the scintillator phi or MRPC strip number, based on the part number.
    If `part < 3`, it is scintillator and the return value is phi number. Otherwise, it is
    MRPC and the return value is strip number.

    Parameters:
        tof_digi_id: The TOF digi ID array or value.
        part: The part number. If not provided, it will be calculated based on the TOF digi ID.

    Returns:
        The scintillator phi or MRPC strip number.
    """
    if part is None:
        return _tof_id_to_phi_or_strip_1(tof_digi_id)
    else:
        return _tof_id_to_phi_or_strip_2(tof_digi_id, part)


@nb.vectorize(cache=True)
def get_tof_digi_id(
    part: IntLike,
    layer_or_module: IntLike,
    phi_or_strip: IntLike,
    end: IntLike,
) -> IntLike:
    """
    Generate TOF scintillator ID based on the part number, layer number, phi number, and readout end number.

    Parameters:
        part: The part number.
        layer_or_module: The scintillator layer or MRPC module number.
        phi_or_strip: The scintillator phi or MRPC strip number.
        end: The readout end number.

    Returns:
        The TOF digi ID.
    """
    if part < 3:
        return np.uint32(
            ((part << DIGI_TOF_PART_OFFSET) & DIGI_TOF_PART_MASK)
            | ((layer_or_module << DIGI_TOF_SCINT_LAYER_OFFSET) & DIGI_TOF_SCINT_LAYER_MASK)
            | ((phi_or_strip << DIGI_TOF_SCINT_PHI_OFFSET) & DIGI_TOF_SCINT_PHI_MASK)
            | ((end << DIGI_TOF_END_OFFSET) & DIGI_TOF_END_MASK)
            | (DIGI_TOF_FLAG << DIGI_FLAG_OFFSET)
        )
    else:
        return np.uint32(
            ((3 << DIGI_TOF_PART_OFFSET) & DIGI_TOF_PART_MASK)
            | (((part - 3) << DIGI_TOF_MRPC_ENDCAP_OFFSET) & DIGI_TOF_MRPC_ENDCAP_MASK)
            | ((layer_or_module << DIGI_TOF_MRPC_MODULE_OFFSET) & DIGI_TOF_MRPC_MODULE_MASK)
            | ((phi_or_strip << DIGI_TOF_MRPC_STRIP_OFFSET) & DIGI_TOF_MRPC_STRIP_MASK)
            | ((end << DIGI_TOF_END_OFFSET) & DIGI_TOF_END_MASK)
            | (DIGI_TOF_FLAG << DIGI_FLAG_OFFSET)
        )


###############################################################################
#                                     EMC                                     #
###############################################################################
@nb.vectorize(cache=True)
def check_emc_id(emc_digi_id: IntLike) -> BoolLike:
    """
    Check if the EMC digi ID is valid.

    Parameters:
        emc_digi_id: The EMC digi ID array or value.

    Returns:
        Whether the digi ID is valid.
    """
    return (emc_digi_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_EMC_FLAG


@nb.vectorize(cache=True)
def emc_id_to_module(emc_digi_id: IntLike) -> IntLike:
    """
    Convert EMC digi ID to module number

    Parameters:
        emc_digi_id: EMC digi ID array or value.

    Returns:
        The module number.
    """
    return np.uint8((emc_digi_id & DIGI_EMC_MODULE_MASK) >> DIGI_EMC_MODULE_OFFSET)


@nb.vectorize(cache=True)
def emc_id_to_theta(emc_digi_id: IntLike) -> IntLike:
    """
    Convert the EMC digi ID to the theta number.

    Parameters:
        emc_digi_id: The EMC digi ID array or value.

    Returns:
        The theta number.
    """
    return np.uint8((emc_digi_id & DIGI_EMC_THETA_MASK) >> DIGI_EMC_THETA_OFFSET)


@nb.vectorize(cache=True)
def emc_id_to_phi(emc_digi_id: IntLike) -> IntLike:
    """
    Convert the EMC digi ID to the phi number.

    Parameters:
        emc_digi_id: The EMC digi ID array or value.

    Returns:
        The phi number.
    """
    return np.uint8((emc_digi_id & DIGI_EMC_PHI_MASK) >> DIGI_EMC_PHI_OFFSET)


@nb.vectorize(cache=True)
def get_emc_digi_id(
    module: IntLike,
    theta: IntLike,
    phi: IntLike,
) -> IntLike:
    """
    Generate EMC digi ID based on the module number, theta number, and phi number.

    Parameters:
        module: The module number.
        theta: The theta number.
        phi: The phi number.

    Returns:
        The EMC digi ID.
    """
    return np.uint32(
        ((module << DIGI_EMC_MODULE_OFFSET) & DIGI_EMC_MODULE_MASK)
        | ((theta << DIGI_EMC_THETA_OFFSET) & DIGI_EMC_THETA_MASK)
        | ((phi << DIGI_EMC_PHI_OFFSET) & DIGI_EMC_PHI_MASK)
        | (DIGI_EMC_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                     MUC                                     #
###############################################################################
@nb.vectorize(cache=True)
def check_muc_id(muc_digi_id: IntLike) -> BoolLike:
    """
    Check if the MUC digi ID is valid.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        Whether the digi ID is valid.
    """
    return (muc_digi_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_MUC_FLAG


@nb.vectorize(cache=True)
def muc_id_to_part(muc_digi_id: IntLike) -> IntLike:
    """
    Convert MUC digi ID to part number

    Parameters:
        muc_digi_id: MUC digi ID array or value.

    Returns:
        The part number.
    """
    return np.uint8((muc_digi_id & DIGI_MUC_PART_MASK) >> DIGI_MUC_PART_OFFSET)


@nb.vectorize(cache=True)
def muc_id_to_segment(muc_digi_id: IntLike) -> IntLike:
    """
    Convert the MUC digi ID to the segment number.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        The segment number.
    """
    return np.uint8((muc_digi_id & DIGI_MUC_SEGMENT_MASK) >> DIGI_MUC_SEGMENT_OFFSET)


@nb.vectorize(cache=True)
def muc_id_to_layer(muc_digi_id: IntLike) -> IntLike:
    """
    Convert the MUC digi ID to the layer number.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        The layer number.
    """
    return np.uint8((muc_digi_id & DIGI_MUC_LAYER_MASK) >> DIGI_MUC_LAYER_OFFSET)


@nb.vectorize(cache=True)
def muc_id_to_channel(muc_digi_id: IntLike) -> IntLike:
    """
    Convert the MUC digi ID to the channel number.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        The channel number.
    """
    return np.uint8((muc_digi_id & DIGI_MUC_CHANNEL_MASK) >> DIGI_MUC_CHANNEL_OFFSET)


@nb.vectorize(cache=True)
def get_muc_digi_id(
    part: IntLike,
    segment: IntLike,
    layer: IntLike,
    channel: IntLike,
) -> IntLike:
    """
    Generate MUC digi ID based on the part number, segment number, layer number, and channel number.

    Parameters:
        part: The part number.
        segment: The segment number.
        layer: The layer number.
        channel: The channel number.

    Returns:
        The MUC digi ID.
    """
    return np.uint32(
        ((part << DIGI_MUC_PART_OFFSET) & DIGI_MUC_PART_MASK)
        | ((segment << DIGI_MUC_SEGMENT_OFFSET) & DIGI_MUC_SEGMENT_MASK)
        | ((layer << DIGI_MUC_LAYER_OFFSET) & DIGI_MUC_LAYER_MASK)
        | ((channel << DIGI_MUC_CHANNEL_OFFSET) & DIGI_MUC_CHANNEL_MASK)
        | (DIGI_MUC_FLAG << DIGI_FLAG_OFFSET)
    )


def muc_id_to_gap(muc_digi_id: IntLike) -> IntLike:
    """
    Convert the MUC digi ID to the gap ID, which is equivalent to layer number.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        The gap ID.
    """
    return muc_id_to_layer(muc_digi_id)


def muc_id_to_strip(muc_digi_id: IntLike) -> IntLike:
    """
    Convert the MUC digi ID to the strip number, which is equivalent to channel number.

    Parameters:
        muc_digi_id: The MUC digi ID array or value.

    Returns:
        The strip number.
    """
    return muc_id_to_channel(muc_digi_id)


###############################################################################
#                                    CGEM                                     #
###############################################################################
@nb.vectorize(cache=True)
def check_cgem_id(cgem_digi_id: IntLike) -> BoolLike:
    """
    Check if the CGEM digi ID is valid.

    Parameters:
        cgem_digi_id: The CGEM digi ID array or value.

    Returns:
        Whether the digi ID is valid.
    """
    return (cgem_digi_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_CGEM_FLAG


@nb.vectorize(cache=True)
def cgem_id_to_layer(cgem_digi_id: IntLike) -> IntLike:
    """
    Convert the CGEM digi ID to the layer number.

    Parameters:
        cgem_digi_id: The CGEM digi ID array or value.

    Returns:
        The layer number.
    """
    return np.uint8((cgem_digi_id & DIGI_CGEM_LAYER_MASK) >> DIGI_CGEM_LAYER_OFFSET)


@nb.vectorize(cache=True)
def cgem_id_to_sheet(cgem_digi_id: IntLike) -> IntLike:
    """
    Convert the CGEM digi ID to the sheet number.

    Parameters:
        cgem_digi_id: The CGEM digi ID array or value.

    Returns:
        The sheet number.
    """
    return np.uint8((cgem_digi_id & DIGI_CGEM_SHEET_MASK) >> DIGI_CGEM_SHEET_OFFSET)


@nb.vectorize(cache=True)
def cgem_id_to_strip(cgem_digi_id: IntLike) -> IntLike:
    """
    Convert CGEM digi ID to strip number

    Parameters:
        cgem_digi_id: CGEM digi ID array or value.

    Returns:
        The strip number.
    """
    return np.uint16((cgem_digi_id & DIGI_CGEM_STRIP_MASK) >> DIGI_CGEM_STRIP_OFFSET)


@nb.vectorize(cache=True)
def cgem_id_to_is_x_strip(cgem_digi_id: IntLike) -> BoolLike:
    """
    Convert the CGEM digi ID to whether it is an X-strip.

    Parameters:
        cgem_digi_id: The CGEM digi ID array or value.

    Returns:
        Whether the strip is an X-strip
    """
    return (
        (cgem_digi_id & DIGI_CGEM_STRIPTYPE_MASK) >> DIGI_CGEM_STRIPTYPE_OFFSET
    ) == DIGI_CGEM_XSTRIP


@nb.vectorize(cache=True)
def get_cgem_digi_id(
    layer: IntLike,
    sheet: IntLike,
    strip: IntLike,
    is_x_strip: BoolLike,
) -> IntLike:
    """
    Generate CGEM digi ID based on the strip number, strip type, sheet number, and layer number.

    Parameters:
        layer: The layer number.
        sheet: The sheet number.
        strip: The strip number.
        is_x_strip: Whether the strip is an X-strip.

    Returns:
        The CGEM digi ID.
    """
    return np.uint32(
        ((strip << DIGI_CGEM_STRIP_OFFSET) & DIGI_CGEM_STRIP_MASK)
        | ((~is_x_strip << DIGI_CGEM_STRIPTYPE_OFFSET) & DIGI_CGEM_STRIPTYPE_MASK)
        | ((sheet << DIGI_CGEM_SHEET_OFFSET) & DIGI_CGEM_SHEET_MASK)
        | ((layer << DIGI_CGEM_LAYER_OFFSET) & DIGI_CGEM_LAYER_MASK)
        | (DIGI_CGEM_FLAG << DIGI_FLAG_OFFSET)
    )



from typing import Any, Optional
import numpy as np
from numpy.typing import NDArray
from uproot_custom.cpp import IElementReader

class Bes3TObjArrayReader(IElementReader):
    def __init__(
        self,
        name: str,
        element_reader: IElementReader,
    ): ...
    def data(self) -> tuple[NDArray[np.uint32], Any]: ...

class Bes3SymMatrixArrayReader(IElementReader):
    def __init__(
        self,
        name: str,
        flat_size: int,
        full_dim: int,
    ): ...
    def data(self) -> NDArray[np.float64]: ...

class Bes3CgemClusterColReader(IElementReader):
    def __init__(self, name: str): ...
    def data(self) -> dict[str, NDArray]: ...

def read_data(
    data: NDArray[np.uint8],
    offsets: NDArray[np.uint32],
    reader: IElementReader,
) -> Any: ...
def read_bes_raw(
    data: NDArray[np.uint32],
    sub_detectors: Optional[list[str]] = None,
) -> dict: ...



from pathlib import Path
from typing import Literal, NamedTuple, Optional

import awkward as ak
import numpy as np
import pytest
import uproot

import pybes3
import pybes3.besio as besio

besio.wrap_uproot()

data_dir = Path(__file__).parent / "data"


class CompareResult(NamedTuple):
    item: str
    is_same: bool
    arr_dtype: Optional[str] = None
    diff_type: Optional[Literal["type mismatch", "length mismatch", "value unequal"]] = None


def compare_ak(arr1: ak.Array, arr2: ak.Array, item_path: str = "") -> list[CompareResult]:
    """
    Compare two awkward arrays and return a list of CompareResult.

    Args:
        arr1 (ak.Array): First array to compare.
        arr2 (ak.Array): Second array to compare.
        item_path (str): Path to the item being compared.

    Returns:
        list[CompareResult]: List of CompareResult objects containing comparison results.
    """
    if arr1.typestr != arr2.typestr:
        return [
            CompareResult(
                item=item_path,
                is_same=False,
                arr_dtype=arr1.typestr,
                diff_type="type mismatch",
            )
        ]

    # Record array
    if arr1.fields:
        res = []
        for field in arr1.fields:
            tmp_res = compare_ak(arr1[field], arr2[field], item_path=f"{item_path}/{field}")

            if (
                len(tmp_res) == 1
                and tmp_res[0].arr_dtype != "ak.Record"
                and tmp_res[0].diff_type == "length mismatch"
            ):
                return [
                    CompareResult(
                        item=item_path + "/*",
                        is_same=False,
                        arr_dtype="ak.Record",
                        diff_type="length mismatch",
                    )
                ]

            else:
                res += tmp_res
        return res

    # Get array dtype
    arr_dtype = arr1.typestr.split("*")[-1].strip()

    try:
        if "float" in arr_dtype and ak.all(np.isclose(arr1, arr2, equal_nan=True)):
            return [
                CompareResult(
                    item=item_path,
                    is_same=True,
                    arr_dtype=arr_dtype,
                )
            ]
        elif ak.all(arr1 == arr2):
            return [
                CompareResult(
                    item=item_path,
                    is_same=True,
                    arr_dtype=arr_dtype,
                )
            ]

        # Value non-equal
        else:
            return [
                CompareResult(
                    item=item_path,
                    is_same=False,
                    diff_type="value unequal",
                    arr_dtype=arr_dtype,
                )
            ]

    # Length mismatch
    except ValueError as e:
        if ak.any(ak.count(arr1, axis=1) != ak.count(arr2, axis=1)):
            return [
                CompareResult(
                    item=item_path,
                    is_same=False,
                    arr_dtype=arr_dtype,
                    diff_type="length mismatch",
                )
            ]
        else:
            print(arr1.typestr)
            print(ak.count(arr1), ak.count(arr2))
            raise e


def test_uproot_branches():
    f_full = uproot.open(data_dir / "test_full_mc_evt_1.rtraw")
    assert len(f_full["Event/TMcEvent"].branches) == 6

    f_only_mc_particles = uproot.open(data_dir / "test_only_mc_particles.rtraw")
    assert len(f_only_mc_particles["Event/TMcEvent"].branches) == 6


def test_mc_full():
    f_rtraw = uproot.open(data_dir / "test_full_mc_evt_1.rtraw")
    truth_arr = ak.from_parquet(data_dir / "test_full_mc_evt_1.rtraw.parquet")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_mc_only_particles():
    f_rtraw = uproot.open(data_dir / "test_only_mc_particles.rtraw")
    truth_arr = ak.from_parquet(data_dir / "test_only_mc_particles.rtraw.parquet")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_dst():
    f_dst = uproot.open(data_dir / "test_full_mc_evt_1.dst")
    truth_arr = ak.from_parquet(data_dir / "test_full_mc_evt_1.dst.parquet")
    arr = f_dst["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_rec():
    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    truth_arr = ak.from_parquet(data_dir / "test_full_mc_evt_1.rec.parquet")
    arr = f_rec["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_cgem_rtraw():
    f_rtraw = uproot.open(data_dir / "test_cgem.rtraw")
    truth_arr = ak.from_parquet(data_dir / "test_cgem.rtraw.parquet")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_cgem_dst():
    f_dst = uproot.open(data_dir / "test_cgem.dst")
    truth_arr = ak.from_parquet(data_dir / "test_cgem.dst.parquet")
    arr = f_dst["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_cgem_rec():
    f_dst = uproot.open(data_dir / "test_cgem.rec")
    truth_arr = ak.from_parquet(data_dir / "test_cgem.rec.parquet")
    arr = f_dst["Event"].arrays()
    assert len(arr) == 10
    comp_res = compare_ak(arr, truth_arr)
    assert all(r.is_same for r in comp_res)


def test_uproot_concatenate():
    arr_concat1 = uproot.concatenate(
        {
            data_dir / "test_full_mc_evt_1.rtraw": "Event",
            data_dir / "test_full_mc_evt_2.rtraw": "Event",
        }
    )
    assert len(arr_concat1) == 20

    arr_concat2 = uproot.concatenate(
        {
            data_dir / "test_full_mc_evt_1.rtraw": "Event/TMcEvent/m_mcParticleCol",
            data_dir / "test_full_mc_evt_2.rtraw": "Event/TMcEvent/m_mcParticleCol",
        }
    )
    assert len(arr_concat2) == 20


def test_bes_open():
    f = besio.open(data_dir / "test_full_mc_evt_1.rtraw")
    assert len(f["Event/TMcEvent"].branches) == 6

    f = besio.open(data_dir / "test_only_mc_particles.rtraw")
    assert len(f["Event/TMcEvent"].branches) == 6


def test_bes_concatenate():
    arr_concat1 = besio.concatenate(
        [data_dir / "test_full_mc_evt_1.rtraw", data_dir / "test_full_mc_evt_2.rtraw"],
        "Event/TMcEvent",
    )
    assert len(arr_concat1) == 20

    arr_concat2 = besio.concatenate(
        [data_dir / "test_full_mc_evt_1.rtraw", data_dir / "test_full_mc_evt_2.rtraw"],
        "Event/TMcEvent/m_mcParticleCol",
    )
    assert len(arr_concat2) == 20


def test_symetric_matrix_expansion():
    def test_symetric_matrix(arr):
        arr = ak.flatten(arr)
        n_dim = int(arr.typestr.split("*")[-2].strip())

        # Check if the matrix is square
        assert n_dim == int(arr.typestr.split("*")[-3].strip())

        for i in range(n_dim):
            for j in range(i, n_dim):
                assert ak.all(arr[:, i, j] == arr[:, j, i])

    f_dst = uproot.open(data_dir / "test_full_mc_evt_1.dst")
    arr_dst = f_dst["Event/TDstEvent"].arrays()

    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    arr_rec = f_rec["Event/TRecEvent"].arrays()

    for tmp_arr in [
        arr_dst.m_mdcTrackCol.m_err,
        arr_dst.m_emcTrackCol.m_err,
        arr_dst.m_extTrackCol.myEmcErrorMatrix,
        arr_dst.m_extTrackCol.myMucErrorMatrix,
        arr_dst.m_extTrackCol.myTof1ErrorMatrix,
        arr_dst.m_extTrackCol.myTof2ErrorMatrix,
        arr_rec.m_recMdcTrackCol.m_err,
        arr_rec.m_recEmcShowerCol.m_err,
        arr_rec.m_recMdcKalTrackCol.m_terror,
    ]:
        test_symetric_matrix(tmp_arr)


def test_digi_expand_TRawData():
    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    arr_digi = f_rec["Event/TDigiEvent"].arrays()
    for field in arr_digi.fields:
        if field == "m_fromMc":
            continue

        assert "TRawData" not in arr_digi[field].fields


@pytest.mark.skipif(
    not Path(
        "/bes3fs/offline/data/raw/round17/231117/run_0079017_All_file001_SFO-1.raw"
    ).exists(),
    reason="Test data is not available",
)
def test_raw():
    f_raw: pybes3.besio.RawBinaryReader = pybes3.open_raw(
        "/bes3fs/offline/data/raw/round17/231117/run_0079017_All_file001_SFO-1.raw"
    )

    n_mdc_digis = ak.Array([1872, 2768, 1641, 2542, 3331, 2672, 2257, 2470, 3635, 3689])

    arr_full = f_raw.arrays(n_blocks=10)
    assert set(arr_full.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert ak.all(ak.count(arr_full.mdc.id, axis=1) == n_mdc_digis)

    arr_mdc = f_raw.arrays(n_blocks=10, sub_detectors=["mdc"])
    assert set(arr_mdc.fields) == {"evt_header", "mdc"}
    assert ak.all(ak.count(arr_mdc.mdc.id, axis=1) == n_mdc_digis)

    arr_batch = f_raw.arrays(n_blocks=10, n_block_per_batch=5)
    assert set(arr_batch.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert ak.all(ak.count(arr_batch.mdc.id, axis=1) == n_mdc_digis)

    arr_workers = f_raw.arrays(n_blocks=10, max_workers=4)
    assert set(arr_workers.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert ak.all(ak.count(arr_workers.mdc.id, axis=1) == n_mdc_digis)


if __name__ == "__main__":
    import pytest

    pytest.main([__file__, "-v", "-s"])



from .emc import (
    emc_barrel_h1,
    emc_barrel_h2,
    emc_barrel_h3,
    emc_barrel_l,
    emc_barrel_offset_1,
    emc_barrel_offset_2,
    emc_barrel_r,
    emc_gid_to_center_x,
    emc_gid_to_center_y,
    emc_gid_to_center_z,
    emc_gid_to_front_center_x,
    emc_gid_to_front_center_y,
    emc_gid_to_front_center_z,
    emc_gid_to_part,
    emc_gid_to_phi,
    emc_gid_to_point_x,
    emc_gid_to_point_y,
    emc_gid_to_point_z,
    emc_gid_to_theta,
    get_emc_crystal_position,
    get_emc_gid,
)
from .mdc import (
    get_mdc_gid,
    get_mdc_wire_position,
    mdc_gid_to_east_x,
    mdc_gid_to_east_y,
    mdc_gid_to_east_z,
    mdc_gid_to_is_stereo,
    mdc_gid_to_layer,
    mdc_gid_to_stereo,
    mdc_gid_to_superlayer,
    mdc_gid_to_west_x,
    mdc_gid_to_west_y,
    mdc_gid_to_west_z,
    mdc_gid_to_wire,
    mdc_gid_z_to_x,
    mdc_gid_z_to_y,
    mdc_layer_to_is_stereo,
    mdc_layer_to_superlayer,
)

__all__ = [
    # EMC
    "emc_barrel_h1",
    "emc_barrel_h2",
    "emc_barrel_h3",
    "emc_barrel_l",
    "emc_barrel_offset_1",
    "emc_barrel_offset_2",
    "emc_barrel_r",
    "emc_gid_to_center_x",
    "emc_gid_to_center_y",
    "emc_gid_to_center_z",
    "emc_gid_to_front_center_x",
    "emc_gid_to_front_center_y",
    "emc_gid_to_front_center_z",
    "emc_gid_to_part",
    "emc_gid_to_phi",
    "emc_gid_to_point_x",
    "emc_gid_to_point_y",
    "emc_gid_to_point_z",
    "emc_gid_to_theta",
    "get_emc_crystal_position",
    "get_emc_gid",
    # MDC
    "get_mdc_gid",
    "get_mdc_wire_position",
    "mdc_gid_to_east_x",
    "mdc_gid_to_east_y",
    "mdc_gid_to_east_z",
    "mdc_gid_to_is_stereo",
    "mdc_gid_to_layer",
    "mdc_gid_to_stereo",
    "mdc_gid_to_superlayer",
    "mdc_gid_to_west_x",
    "mdc_gid_to_west_y",
    "mdc_gid_to_west_z",
    "mdc_gid_to_wire",
    "mdc_gid_z_to_x",
    "mdc_gid_z_to_y",
    "mdc_layer_to_is_stereo",
    "mdc_layer_to_superlayer",
]



import numpy as np

import pybes3 as p3
import pybes3.detectors.geometry as geom


def test_mdc_geom():
    gid: np.ndarray = p3.get_mdc_wire_position()["gid"]
    assert np.all(p3.get_mdc_gid(geom.mdc._layer, geom.mdc._wire) == gid)
    assert np.all(p3.mdc_gid_to_superlayer(gid) == geom.mdc._superlayer)
    assert np.all(p3.mdc_layer_to_superlayer(geom.mdc._layer) == geom.mdc._superlayer)
    assert np.all(p3.mdc_gid_to_layer(gid) == geom.mdc._layer)
    assert np.all(p3.mdc_gid_to_wire(gid) == geom.mdc._wire)
    assert np.all(p3.mdc_gid_to_stereo(gid) == geom.mdc._stereo)
    assert np.all(p3.mdc_layer_to_is_stereo(geom.mdc._layer) == geom.mdc._is_stereo)
    assert np.all(p3.mdc_gid_to_is_stereo(gid) == geom.mdc._is_stereo)
    assert np.all(p3.mdc_gid_to_east_x(gid) == geom.mdc._east_x)
    assert np.all(p3.mdc_gid_to_east_y(gid) == geom.mdc._east_y)
    assert np.all(p3.mdc_gid_to_east_z(gid) == geom.mdc._east_z)
    assert np.all(p3.mdc_gid_to_west_x(gid) == geom.mdc._west_x)
    assert np.all(p3.mdc_gid_to_west_y(gid) == geom.mdc._west_y)
    assert np.all(p3.mdc_gid_to_west_z(gid) == geom.mdc._west_z)

    assert np.allclose(p3.mdc_gid_z_to_x(gid, geom.mdc._west_z), geom.mdc._west_x, atol=1e-6)
    assert np.allclose(p3.mdc_gid_z_to_y(gid, geom.mdc._west_z), geom.mdc._west_y, atol=1e-6)
    assert np.allclose(p3.mdc_gid_z_to_x(gid, geom.mdc._east_z), geom.mdc._east_x, atol=1e-6)
    assert np.allclose(p3.mdc_gid_z_to_y(gid, geom.mdc._east_z), geom.mdc._east_y, atol=1e-6)


def test_emc_geom():
    gid: np.ndarray = p3.get_emc_crystal_position()["gid"]
    assert np.all(p3.get_emc_gid(geom.emc._part, geom.emc._theta, geom.emc._phi) == gid)
    assert np.all(p3.emc_gid_to_part(gid) == geom.emc._part)
    assert np.all(p3.emc_gid_to_theta(gid) == geom.emc._theta)
    assert np.all(p3.emc_gid_to_phi(gid) == geom.emc._phi)

    for i in range(8):
        assert np.all(p3.emc_gid_to_point_x(gid, i) == geom.emc._points_x[gid, i])
        assert np.all(p3.emc_gid_to_point_y(gid, i) == geom.emc._points_y[gid, i])
        assert np.all(p3.emc_gid_to_point_z(gid, i) == geom.emc._points_z[gid, i])

    assert np.allclose(p3.emc_gid_to_center_x(gid), geom.emc._center_x, atol=1e-6)
    assert np.allclose(p3.emc_gid_to_center_y(gid), geom.emc._center_y, atol=1e-6)
    assert np.allclose(p3.emc_gid_to_center_z(gid), geom.emc._center_z, atol=1e-6)
    assert np.allclose(p3.emc_gid_to_front_center_x(gid), geom.emc._front_center_x, atol=1e-6)
    assert np.allclose(p3.emc_gid_to_front_center_y(gid), geom.emc._front_center_y, atol=1e-6)
    assert np.allclose(p3.emc_gid_to_front_center_z(gid), geom.emc._front_center_z, atol=1e-6)


if __name__ == "__main__":
    import pytest

    pytest.main([__file__, "-v", "-s"])
